#include "CommonUtils/TabInterpolator.hpp"
#include "CommonUtils/FileCommentRemover.hpp"
#include "CommonUtils/TabAdjuster.hpp"
#include "CommonUtils/DigitAdjuster.hpp"
#include "NuclearDataUtils/OtherUtils/Endf6FileStore.hpp"
#include "EndfUtils/Endf6Parser/Endf6Parser.hpp"
#include "TabUtils/IntegrateTabData.hpp"
#include "TabUtils/LinearizeTabData.hpp"

using namespace frendy;

////////////////////////////////////////////////////////////////////////////////////////////////////

void read_input(string inp_name, string& out_name, string& ref_name, string& comp_name)
{
  FileCommentRemover rc_obj;
  string inp_name_mod = inp_name + "aaabbbcccdddeeefffggg.hhh";

  rc_obj.remove_comment(inp_name, inp_name_mod);

  ifstream fin;
  fin.open(inp_name_mod.c_str());
  fin >> out_name;
  fin >> ref_name;
  fin >> comp_name;
  cout << "Output file name            : " << out_name  << endl;
  cout << "Reference PENDF file name   : " << ref_name  << endl;
  cout << "Compare PENDF file name     : " << comp_name << endl;
  fin.close();
  remove(inp_name_mod.c_str());
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void read_mt_list( string file_name, vector<Integer>& mt_list )
{
  string mod_name = file_name + "aaabbbcccdddeeefff";

  FileCommentRemover remove_obj;
  remove_obj.remove_comment(file_name, mod_name);

  ifstream fin;
  fin.open(mod_name.c_str());

  int mt_val = 0;
  while( !fin.eof() )
  {
    fin >> mt_val;
    if( mt_val > 0 && mt_val < 1000 )
    {
      mt_list.push_back(mt_val);
      mt_val = 0;
    }
  }

  fin.close();
  remove(mod_name.c_str());
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void set_xs_type_list(string file_name,  vector<Integer>& xs_type_list)
{
  //Allocate Endf6Parser class to endf_obj;
  Endf6Parser endf_obj;

  //Set pendf file name
  endf_obj.set_file_name_xs_only(file_name);

  //Get MF3 data
  vector<MF03Parser> mf03 = endf_obj.get_mf03_data();
  endf_obj.clear();

  xs_type_list.clear();
  for(int i=0; i<static_cast<int>(mf03.size()); i++)
  {
    Integer mt_no = mf03[i].get_mt_no();
    if( mt_no != unassigned_mt_no )
    {
      xs_type_list.push_back(mt_no);
    }
  }
}


////////////////////////////////////////////////////////////////////////////////////////////////////

void check_xs_type_list(vector<Integer>  xs_type_list, vector<Integer> xs_type_list_comp,
                        vector<Integer>& xs_type_list_mod)
{
  xs_type_list_mod.clear();

  int i_max = static_cast<int>(xs_type_list.size());
  int j_max = static_cast<int>(xs_type_list_comp.size());
  for(int i=0; i<i_max; i++)
  {
    Integer react_no = xs_type_list[i];

    int react_flg = 0;
    for(int j=0; j<j_max; j++)
    {
      if( xs_type_list_comp[j] == react_no )
      {
        react_flg = 1;
        xs_type_list_mod.push_back(react_no);
        break;
      }
    }

    if( react_flg == 0 )
    {
      ErrorManager err_obj;

      string class_name = "comp_xs_line.cpp";
      string func_name  = "check_xs_type_list";

      ostringstream oss;
      oss << react_no;
      string str_data = "Reaction type : " + oss.str();

      vector<string> err_com;
      err_com.push_back(str_data);
      err_com.push_back("This reaction type is not found in comparison results.");
      err_com.push_back("Comparison of this reaction type is skipped.");

      err_obj.output_caution(class_name, func_name, err_com);
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void set_xs_data(string file_name,  vector<Integer> xs_type_list,
                 vector<vector<Integer> >& nbt_xs,   vector<vector<Integer> >& int_xs,
                 vector<vector<Real> >&    e_int_xs, vector<vector<Real> >&    xs_tab)
{
  //Allocate Endf6Parser class to endf_obj;
  Endf6Parser endf_obj;

  //Set pendf file name
  endf_obj.set_file_name_xs_only(file_name);

  //Resize xs data
  int xs_data_no = static_cast<int>(xs_type_list.size());
  nbt_xs.resize(xs_data_no);
  int_xs.resize(xs_data_no);
  e_int_xs.resize(xs_data_no);
  xs_tab.resize(xs_data_no);

  //Linearizing TAB1 data
  LinearizeTabData  lin_tab_obj;

  //Get MF3 data
  vector<MF03Parser> mf03 = endf_obj.get_mf03_data();
  for(int i=0; i<static_cast<int>(mf03.size()); i++)
  {
    Integer mt_no = mf03[i].get_mt_no();
    for(int j=0; j<xs_data_no; j++)
    {
      if( mt_no == xs_type_list[j] )
      {
        nbt_xs[j]   = mf03[i].get_NBT_SIG();
        int_xs[j]   = mf03[i].get_INT_SIG();
        e_int_xs[j] = mf03[i].get_E_INT_SIG();
        xs_tab[j]   = mf03[i].get_SIG_TAB();

        lin_tab_obj.linearize_tab1_data(nbt_xs[j], int_xs[j], e_int_xs[j], xs_tab[j]);
        break;
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void read_weight_data(string weight_file_name, vector<Real> e_int_weight, vector<Real>& weight_tab)
{
  FileCommentRemover file_mod_obj;
  vector<string> input_vec = file_mod_obj.remove_comment(weight_file_name);
  file_mod_obj.clear();

  StringUtils   str_obj;
  VectorClearer clr_obj;
  str_obj.set_text_data(input_vec);
  clr_obj.clear_vec_array1_str(input_vec);

  vector<vector<string> > input_data = str_obj.get_div_string_vec_all();
  str_obj.clear();

  vector<Real> e_weight_ori, weight_tab_ori;
  e_weight_ori.clear();
  weight_tab_ori.clear();

  int i_max  = static_cast<int>(input_data.size());
  int ele_no = 0;
  for(int i=0; i<i_max; i++)
  {
    int j_max = static_cast<int>(input_data[i].size());
    for(int j=0; j<j_max; j++)
    {
      if( static_cast<int>(input_data[i][j].length()) > 0 )
      {
        if( ele_no%2 == 0 )
        {
          e_weight_ori.push_back(numeric_string_cast<Real>(input_data[i][j]));
        }
        else //if( ele_no%2 == 1 )
        {
          weight_tab_ori.push_back(numeric_string_cast<Real>(input_data[i][j]));
        }
        ele_no++;
      }
    }
  }
  if( ele_no %2 == 1 )
  {
    int pos_no = static_cast<int>(weight_tab_ori.size());
    weight_tab_ori.push_back(weight_tab_ori[pos_no-1]);
  }

  vector<Integer> nbt_vec, int_vec;
  nbt_vec.resize(1);
  int_vec.resize(1);
  nbt_vec[0] = static_cast<int>(e_weight_ori.size());
  int_vec[0] = int_lin_lin;

  TabInterpolator  ti_obj;

  int ene_no = static_cast<int>(e_int_weight.size());
  weight_tab.resize(ene_no);
  for(int i=0; i<ene_no; i++)
  {
    weight_tab[i] = 0.0;
    ti_obj.interpolation_tab1(e_int_weight[i], weight_tab[i],
                              nbt_vec, int_vec, e_weight_ori, weight_tab_ori);
  }
  nbt_vec.clear();
  int_vec.clear();
  e_weight_ori.clear();
  weight_tab_ori.clear();
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void set_weight(string weight_mode, vector<Real>& e_vec_ori, vector<Real>& weight_tab)
{
  int ene_no = static_cast<int>(e_vec_ori.size());

  weight_tab.resize(ene_no);
  if( weight_mode == "const" )
  {
    for(int i=0; i<ene_no; i++)
    {
      weight_tab[i] = 1.0;
    }
  }
  else if( weight_mode == "1/e" )
  {
    for(int i=0; i<ene_no; i++)
    {
      weight_tab[i] = 1.0 / e_vec_ori[i];
    }
  }
  else
  {
    read_weight_data(weight_mode, e_vec_ori, weight_tab);
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void multi_group_generator(string mg_file_name, string weight_mode,
                           vector<vector<Integer> >& nbt_xs,   vector<vector<Integer> >& int_xs,
                           vector<vector<Real> >&    e_int_xs, vector<vector<Real> >&    xs_tab)
{
  ifstream fin(mg_file_name.c_str());
  vector<Real> mg_grid, mg_el, mg_eh, mg_ref;

  if(!fin)
  {
    cout << endl;
    cout << "Multi-group structure data file : " << mg_file_name << endl;
    cout << "Multi-group structure data file cannot be found." << endl;
    throw runtime_error("Please check multi-group structure data file.");
  }

  while(!fin.eof())
  {
    Real grid_data;
    fin >> grid_data;
    mg_grid.push_back(grid_data);
  }

  TabAdjuster ad_obj;
  ad_obj.delete_overlap_grid(mg_grid);

  cout.precision(7);
  cout.setf(ios::showpoint);
  cout.setf(ios::scientific);
  //cout << "    Energy grid structure [MeV]" << endl;
  int unit_check = 0; //MeV
  for(int i=0; i<static_cast<int>(mg_grid.size()); i++)
  {
    if( mg_grid[i] < 1.0E-10 )
    {
      break;
    }
    else if( mg_grid[i] > 1.0E+6 )
    {
      unit_check = 1; //eV
      break;
    }
  }
  if( unit_check == 0 )
  {
    cout << "The unit of multigroup energy group structure is modified from MeV to eV." << endl;
    for(int i=0; i<static_cast<int>(mg_grid.size()); i++)
    {
      //cout << "      " << i+1 << "\t" << mg_grid[i] << endl;
      mg_grid[i] *= 1.0E+6; //[MeV] -> [eV]
    }
  }

  DigitAdjuster   digit_obj;
  int mg_no = static_cast<int>(mg_grid.size()) - 1;
  mg_ref.resize(mg_no*2);
  mg_el.resize(mg_no);
  mg_eh.resize(mg_no);
  for(int i=0; i<mg_no; i++)
  {
    mg_ref[2*i]   = digit_obj.get_adjusted_value(mg_grid[i],   sig_fig, -1.0); //el of each energy grid
    mg_ref[2*i+1] = digit_obj.get_adjusted_value(mg_grid[i+1], sig_fig, +1.0); //eh of each energy grid
  }

  TabInterpolator  ti_obj;
  IntegrateTabData integ_obj;
  int i_max = static_cast<int>(xs_tab.size());
  for(int i=0; i<i_max; i++)
  {
    if(static_cast<int>(xs_tab[i].size()) == 0 )
    {
      nbt_xs[i].clear();
      int_xs[i].clear();
      e_int_xs[i].clear();
      xs_tab[i].clear();
      continue;
    }

    vector<Real> weight_tab;
    set_weight(weight_mode, e_int_xs[i], weight_tab);

    vector<Real> xs_tab_mod = xs_tab[i];
    int j_max = static_cast<int>(xs_tab_mod.size());
    for(int j=0; j<j_max; j++)
    {
      xs_tab_mod[j] *= weight_tab[j];
    }

    vector<Integer> nbt_weight, int_weight;
    nbt_weight.resize(1);
    int_weight.resize(1);
    nbt_weight[0] = j_max;
    int_weight[0] = int_lin_lin;

    Real ene_min = e_int_xs[i][0];
    Real ene_max = e_int_xs[i][static_cast<int>(e_int_xs[i].size())-1];

    vector<Real> xs_tab_mg;
    xs_tab_mg.resize(mg_no*2);
    for(int j=0; j<mg_no; j++)
    {
      Real xs_mg_val     = 0.0;
      Real weight_mg_val = 0.0;
      Real el            = mg_grid[j];
      Real eh            = mg_grid[j+1];
      if( ene_min > eh || ene_max < el )
      {
        xs_mg_val = 0.0;
      }
      else
      {
        if( ene_min > el )
        {
          el = ene_min;
        }

        if( ene_max < eh )
        {
          eh = ene_max;
        }
        xs_mg_val = integ_obj.integrate_tab1_data(nbt_xs[i], int_xs[i], e_int_xs[i], xs_tab_mod, el, eh);
        xs_mg_val /= (eh - el);

        weight_mg_val = integ_obj.integrate_tab1_data(nbt_weight, int_weight, e_int_xs[i], weight_tab, el, eh);
        weight_mg_val /= (eh - el);

        if( fabs(weight_mg_val) > min_value )
        {
          xs_mg_val /= weight_mg_val;
        }
      }
      xs_tab_mg[2*j]   = xs_mg_val;
      xs_tab_mg[2*j+1] = xs_mg_val;
    }

    xs_tab[i]   = xs_tab_mg;
    e_int_xs[i] = mg_ref;
    nbt_xs[i].resize(1);
    int_xs[i].resize(1);
    nbt_xs[i][0] = mg_no;
    int_xs[i][0] = int_lin_lin;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void modify_xs_data(vector<vector<Real> >& e_ref,
                    vector<vector<Integer> >& nbt_xs,   vector<vector<Integer> >& int_xs,
                    vector<vector<Real> >&    e_int_xs, vector<vector<Real> >&    xs_tab)
{
  TabInterpolator ti_obj;

  int i_max = static_cast<int>(e_ref.size());
  for(int i=0; i<i_max; i++)
  {
    vector<Integer> nbt_mod, int_mod;
    vector<Real>    xs_mod;

    if( static_cast<int>(xs_tab[i].size()) == 0 )
    {
      nbt_xs[i].clear();
      int_xs[i].clear();
      e_int_xs[i].clear();
      xs_tab[i].clear();
      continue;
    }

    int j_max = static_cast<int>(e_ref[i].size());
    nbt_mod.push_back(j_max);
    int_mod.push_back(int_lin_lin);

    xs_mod.resize(j_max);
    for(int j=0; j<j_max; j++)
    {
      ti_obj.interpolation_tab1(e_ref[i][j], xs_mod[j], nbt_xs[i], int_xs[i], e_int_xs[i], xs_tab[i]);
    }
    nbt_xs[i]   = nbt_mod;
    int_xs[i]   = int_mod;
    e_int_xs[i] = e_ref[i];
    xs_tab[i]   = xs_mod;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void comp_xs(string ref_name, string comp_name, string mg_structure_file_name, string weight_mode,
             vector<Integer> xs_type_list, vector<vector<Real> >& ene_vec,
             vector<vector<Real> >& xs_ref_vec, vector<vector<Real> >& xs_comp_vec,
             vector<vector<Real> >& xs_dif_vec, vector<Real>& rms_vec)
{
  vector<vector<Integer> > nbt_xs_ref,   int_xs_ref, nbt_xs_comp, int_xs_comp;
  vector<vector<Real> >    e_int_xs_ref, xs_tab_ref, e_int_xs_comp, xs_tab_comp;
  set_xs_data(ref_name,  xs_type_list, nbt_xs_ref,  int_xs_ref,  e_int_xs_ref,  xs_tab_ref);
  set_xs_data(comp_name, xs_type_list, nbt_xs_comp, int_xs_comp, e_int_xs_comp, xs_tab_comp);

  if( static_cast<int>(mg_structure_file_name.size()) > 0 )
  {
    multi_group_generator(mg_structure_file_name, weight_mode,
                          nbt_xs_ref,  int_xs_ref,  e_int_xs_ref,  xs_tab_ref);
    multi_group_generator(mg_structure_file_name, weight_mode,
                          nbt_xs_comp, int_xs_comp, e_int_xs_comp, xs_tab_comp);
  }
  else
  {
    modify_xs_data(e_int_xs_ref, nbt_xs_comp, int_xs_comp, e_int_xs_comp, xs_tab_comp);
  }

  ene_vec     = e_int_xs_ref;
  xs_ref_vec  = xs_tab_ref;
  xs_comp_vec = xs_tab_comp;
  
  int xs_data_no = static_cast<int>(xs_type_list.size());

  xs_dif_vec.clear();
  rms_vec.clear();
  xs_dif_vec.resize(xs_data_no);
  rms_vec.resize(xs_data_no);
  for(int i=0; i<xs_data_no; i++)
  {
    rms_vec[i] = 0.0;

    int j_max = static_cast<int>(e_int_xs_ref[i].size());
    xs_dif_vec[i].resize(j_max);
    for(int j=0; j<j_max; j++)
    {
      if( fabs(xs_tab_ref[i][j]) > min_value )
      {
        xs_dif_vec[i][j] = (xs_comp_vec[i][j] - xs_tab_ref[i][j]) / xs_tab_ref[i][j];
      }
      else
      {
        xs_dif_vec[i][j] = (xs_comp_vec[i][j] - xs_tab_ref[i][j]);
      }

      rms_vec[i] += xs_dif_vec[i][j] * xs_dif_vec[i][j];
    }

    if( j_max > 0 )
    {
      rms_vec[i] = sqrt(rms_vec[i] / static_cast<Real>(j_max));
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void output_result(string out_name, string ref_name, string comp_name,
                   string mg_structure_file_name, string mg_weight_opt,
                   vector<Integer> xs_type_list, vector<vector<Real> >& ene_vec,
                   vector<vector<Real> >& xs_ref_vec, vector<vector<Real> >& xs_comp_vec,
                   vector<vector<Real> >& xs_dif_vec)
{
  int xs_data_no = static_cast<int>(xs_type_list.size());
  for(int i=0; i<xs_data_no; i++)
  {
    int j_max = static_cast<int>(ene_vec[i].size());
    if( j_max == 0 )
    {
      continue;
    }

    ostringstream  oss;
    oss << xs_type_list[i];

    string out_name_mod = out_name + "_mt";
    if( xs_type_list[i] < 10 )
    {
      out_name_mod = out_name_mod + "00";
    }
    else if( xs_type_list[i] < 100 )
    {
      out_name_mod = out_name_mod + "0";
    }
    out_name_mod = out_name_mod + oss.str() + ".out";

    ofstream fout;
    fout.open(out_name_mod.c_str());

    fout << "# comp data             : " << comp_name << endl;
    fout << "# ref  data             : " << ref_name  << endl;
    if( static_cast<int>(mg_structure_file_name.size()) > 0 )
    {
      fout << "# multi group structure : " << mg_structure_file_name << endl;
      fout << "# weight mode           : " << mg_weight_opt << endl;
    }
    fout << "#" << endl;
    fout << "# energy [eV]   / comp data       / ref data        / relative dif (comp - ref)/ref" << endl;

    fout.precision(10);
    fout.setf(ios::showpoint);
    fout.setf(ios::scientific);

    for(int j=0; j<j_max; j++)
    {
      fout << ene_vec[i][j] << "  "
           << xs_comp_vec[i][j] << "  " << xs_ref_vec[i][j] << "  " << xs_dif_vec[i][j] << endl;
    }
    fout.close();
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[])
{
  ErrorManager err_obj;
  //err_obj.set_err_mes_opt(err_obj.err_mes_debug);
  err_obj.set_err_mes_opt(err_obj.err_mes_default);
  //err_obj.set_err_mes_opt(err_obj.err_mes_skip_warning);

  if( argc <= 1 )
  {
    throw runtime_error("Please set input file name.");
  }

  string inp_name = argv[1];

  //Read input data
  string          out_name, ref_name, comp_name;
  read_input(inp_name, out_name, ref_name, comp_name);

  //Read MT number list
  vector<Integer> xs_type_list;
  xs_type_list.clear();
  if( argc > 2 )
  {
    string mt_list_name = argv[2];

    string mt_list_check = mt_list_name;
    transform(mt_list_check.begin(), mt_list_check.end(), mt_list_check.begin(), ::tolower);
    if( mt_list_check == "none" || mt_list_check == "off" || mt_list_check == "skip" )
    {
      cout << "This program compares all mt data." << endl;
    }
    else
    {
      read_mt_list(mt_list_name, xs_type_list);

      if( static_cast<int>(xs_type_list.size()) == 0 )
      {
        cout << "Though user sets mt-list file, there is no mt data." << endl;
        cout << "This program compares all mt data." << endl;
      }
    }
  }

  string mg_structure_file_name = "";
  string weight_mode            = "";

  int arg_no = 3;
  if( argc > arg_no )
  {
    mg_structure_file_name = argv[arg_no];

    weight_mode = "const";
    if( argc > arg_no+1 )
    {
      string mg_weight_file_name = argv[arg_no+1];
      string mg_weight_opt       = mg_weight_file_name;
      transform(mg_weight_opt.begin(), mg_weight_opt.end(), mg_weight_opt.begin(), ::tolower);
      if( mg_weight_opt == "const" )
      {
        weight_mode = "const";
      }
      else if( mg_weight_opt == "1/e" )
      {
        weight_mode = "1/e";
      }
      else
      {
        weight_mode = mg_weight_file_name;
      }
    }

    cout << "  multi-group generation using " << mg_structure_file_name << endl;
    cout << "  weighting data                " << weight_mode           << endl;
  }

  if( static_cast<int>(xs_type_list.size()) == 0 )
  {
    //Read mt list from PENDF file
    vector<Integer> xs_type_list_comp, xs_type_list_mod;
    set_xs_type_list(ref_name,  xs_type_list);
    set_xs_type_list(comp_name, xs_type_list_comp);
    check_xs_type_list(xs_type_list, xs_type_list_comp, xs_type_list_mod);
    xs_type_list = xs_type_list_mod;
  }

  //Read cross section data and compare these data
  vector<vector<Real> > ene_vec, xs_ref_vec, xs_comp_vec, xs_dif_vec;
  vector<Real>          rms_vec;
  comp_xs(ref_name, comp_name, mg_structure_file_name, weight_mode, xs_type_list,
          ene_vec, xs_ref_vec, xs_comp_vec, xs_dif_vec, rms_vec);

  //Output comparison results
  output_result(out_name, ref_name, comp_name, mg_structure_file_name, weight_mode,
                xs_type_list, ene_vec, xs_ref_vec, xs_comp_vec, xs_dif_vec);


  int xs_data_no = static_cast<int>(xs_type_list.size());

  cout.precision(10);
  cout.setf(ios::scientific);
  cout << "Reaction type\tEnergy grid number\tRMS value [%]" << endl;
  for(int i=0; i<xs_data_no; i++)
  {
    if( static_cast<int>(ene_vec[i].size()) > 0 )
    {
      cout << xs_type_list[i] << "\t" << ene_vec[i].size() << "\t" << rms_vec[i]*100.0 << endl;
    }
  }

  return 0;
}

