#include "CommonUtils/TabInterpolator.hpp"
#include "NuclearDataUtils/OtherUtils/Endf6FileStore.hpp"
#include "ContinuousEnergyXSUtils/AceDataParser/FastAceDataParser.hpp"
#include "CommonUtils/FileCommentRemover.hpp"

using namespace frendy;

////////////////////////////////////////////////////////////////////////////////////////////////////

void read_mt_list( string file_name, vector<int>& mt_list )
{
  string mod_name = file_name + "aaabbbcccdddeeefff";

  FileCommentRemover remove_obj;
  remove_obj.remove_comment(file_name, mod_name);

  ifstream fin;
  fin.open(mod_name.c_str());

  int mt_val = 0;
  while( !fin.eof() )
  {
    fin >> mt_val;
    if( mt_val > 0 && mt_val < 1000 )
    {
      mt_list.push_back(mt_val);
    }
    else if( mt_val == 1018 ||
             mt_val == 1452 || mt_val == 1455 || mt_val == 1456 ||
             mt_val == 2452 || mt_val == 2455 || mt_val == 2456 )
    {
      mt_list.push_back(mt_val);
    }
    mt_val = 0;
  }

  fin.close();
  remove(mod_name.c_str());
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void read_nu_data(FastAceDataObject& ace_obj, int mt_no, vector<Real>& e_vec, vector<Real>& nu_vec)
{
  ace_obj.get_xs_data(1, e_vec, nu_vec);
  nu_vec.clear();

  vector<Real> nu_coef;
  if( mt_no == 1452 )
  {
    nu_coef = ace_obj.get_nu_data();
  }
  else if( mt_no == 1456 )
  {
    nu_coef = ace_obj.get_nu_p_data();
  }
  else if( mt_no == 1455 )
  {
    nu_coef = ace_obj.get_nu_d_data();
  }
  else
  {
    return;
  }

  int i_max = static_cast<int>(e_vec.size());
  int j_max = static_cast<int>(nu_coef.size());
  if( j_max == 0 )
  {
    e_vec.clear();
    nu_vec.clear();
    return;
  }

  nu_vec.resize(i_max);
  for(int i=0; i<i_max; i++)
  {
    Real e_coef = 1.0;
    nu_vec[i]   = 0.0;
    for(int j=0; j<j_max; j++)
    {
      nu_vec[i] += nu_coef[j] * e_coef;
      e_coef    *= e_vec[i];
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void read_ace_file(string file_name, vector<int>& mt_list,
                   vector<vector<Integer> >& nbt_xs,   vector<vector<Integer> >& int_xs,
                   vector<vector<Real> >&    e_int_xs, vector<vector<Real> >&    xs_tab)
{
  //Read ACE file
  FastAceDataParser read_obj;
  read_obj.set_file_name(file_name);
  FastAceDataObject ace_obj = read_obj.get_ace_data_obj();

  int i_max = static_cast<int>(mt_list.size());
  if( i_max == 0 )
  {
    mt_list.push_back(  1); //total
    mt_list.push_back(  2); //scatter;
    mt_list.push_back(101); //disappearance

    int fis_flg = 0;
    vector<int> mt_list_add = ace_obj.get_react_type_list();
    for(int i=0; i<static_cast<int>(mt_list_add.size()); i++)
    {
      mt_list.push_back(mt_list_add[i]);
      if( mt_list_add[i] == 18 )
      {
        fis_flg = 1;
      }
    }
    mt_list_add.clear();

    if( fis_flg > 0 )
    {
      //For chi (fission spectrum)
      mt_list.push_back(1018);

      //For nu_tot (total number of neutrons per fission)
      if( static_cast<int>(ace_obj.get_nu_data().size()) > 0 )
      {
        int mt_add = static_cast<int>(ace_obj.get_nu_bar_repr_flg())*1000 + 452;
        mt_list.push_back(mt_add);
      }

      //For nu_f
      if( static_cast<int>(ace_obj.get_nu_d_data().size()) > 0 &&
          static_cast<int>(ace_obj.get_nu_d_bar_repr_flg()) == 2 )
      {
        int mt_add = static_cast<int>(ace_obj.get_nu_d_bar_repr_flg())*1000 + 455;
        mt_list.push_back(mt_add);
      }

      //For nu_p
      if( static_cast<int>(ace_obj.get_nu_p_data().size()) > 0 )
      {
        int mt_add = static_cast<int>(ace_obj.get_nu_p_bar_repr_flg())*1000 + 456;
        mt_list.push_back(mt_add);
      }
    }
    
    i_max   = static_cast<int>(mt_list.size());
  }

  //Resize xs data
  nbt_xs.resize(i_max);
  int_xs.resize(i_max);
  e_int_xs.resize(i_max);
  xs_tab.resize(i_max);
  
  for(int i=0; i<i_max; i++)
  {
    if( mt_list[i] < 1000 )
    {
      ace_obj.get_xs_data(mt_list[i], e_int_xs[i], xs_tab[i]);
    }
    else
    {
      if( mt_list[i] == 1018 )
      {
        ace_obj.get_chi_data(e_int_xs[i], xs_tab[i]);
      }
      else if( mt_list[i] < 2000 )
      {
        read_nu_data(ace_obj, mt_list[i], e_int_xs[i], xs_tab[i]);
      }
      else if( mt_list[i] == 2452 )
      {
        e_int_xs[i] = ace_obj.get_nu_ene_data();
        xs_tab[i]   = ace_obj.get_nu_data();
      }
      else if( mt_list[i] == 2455 )
      {
        e_int_xs[i] = ace_obj.get_nu_d_ene_data();
        xs_tab[i]   = ace_obj.get_nu_d_data();
      }
      else if( mt_list[i] == 2456 )
      {
        e_int_xs[i] = ace_obj.get_nu_p_ene_data();
        xs_tab[i]   = ace_obj.get_nu_p_data();
      }
    }

    int j_max = static_cast<int>(e_int_xs[i].size());
    nbt_xs[i].resize(1);
    int_xs[i].resize(1);
    nbt_xs[i][0] = j_max;
    int_xs[i][0] = int_lin_lin;

    for(int j=0; j<j_max; j++)
    {
      e_int_xs[i][j] *= 1.0E6; //MeV -> eV
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void output_result(string out_name, vector<Integer> xs_type_list, 
                   vector<vector<Real> >& ene_vec, vector<vector<Real> >& xs_vec)
{
  int xs_data_no = static_cast<int>(xs_type_list.size());
  for(int i=0; i<xs_data_no; i++)
  {
    ostringstream  oss;
    oss << xs_type_list[i];

    string out_name_mod = out_name + "_mt";
    if( xs_type_list[i] < 10 )
    {
      out_name_mod = out_name_mod + "00";
    }
    else if( xs_type_list[i] < 100 )
    {
      out_name_mod = out_name_mod + "0";
    }
    out_name_mod = out_name_mod + oss.str() + ".xs";

    int j_max = static_cast<int>(ene_vec[i].size());
    if( j_max > 0 )
    {
      ofstream fout;
      fout.open(out_name_mod.c_str());

      fout.precision(10);
      fout.setf(ios::showpoint);
      fout.setf(ios::scientific);

      for(int j=0; j<j_max; j++)
      {
        fout << ene_vec[i][j] << "  " << xs_vec[i][j] << endl;
      }
      fout.close();
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[])
{
  if( argc <= 1 )
  {
    throw runtime_error("Please set ace file name.");
  }

  string  ace_file_name = argv[1];

  vector<int> mt_list;
  mt_list.clear();
//  mt_list.push_back(  1); //total
//  mt_list.push_back(  2); //scatter;
//  mt_list.push_back(101); //disappearance
  if( argc > 2 )
  {
    string mt_list_name = argv[2];
    read_mt_list(mt_list_name, mt_list);

    if( static_cast<int>(mt_list.size()) == 0 )
    {
      cout << "Though user sets mt-list file, there is no mt data." << endl;
      cout << "This program compares in all mt data." << endl;
    }
  }

  vector<vector<Integer> > nbt_xs,   int_xs;
  vector<vector<Real> >    e_int_xs, xs_tab;

  cout << "  read_ace_file " << ace_file_name  << endl;
  read_ace_file(ace_file_name,  mt_list,
                nbt_xs, int_xs, e_int_xs, xs_tab);

  //output ene, xs_r, xs_c
  cout << "  output_results" << endl;
  output_result(ace_file_name, mt_list, e_int_xs, xs_tab);
}

